import pytest
import sys

sys.path.append('../')
from pynyzo.messages.statusresponse import StatusResponse
from pynyzo.messages.blockrequest import BlockRequest


def test_status_1(verbose=False):
    strings = ['line1', 'line2']
    message = StatusResponse(lines=strings)
    if verbose:
        print(message.get_bytes())
    assert message.get_bytes() == b'\x02\x00\x05line1\x00\x05line2'
    message2 = StatusResponse(buffer=b'0000000000' + message.get_bytes())  # 10 more bytes to account for timestamp + type
    if verbose:
        print(message2.get_lines())
    assert str(message2.get_lines()) == "['line1', 'line2']"


def test_block_request(verbose=False):
    message = BlockRequest(start_height=10, end_height=20, include_balance_list=True)
    if verbose:
        print(message.get_bytes())
    assert message.get_bytes() == b'\x00\x00\x00\x00\x00\x00\x00\n\x00\x00\x00\x00\x00\x00\x00\x14\x01'
    message2 = BlockRequest(buffer=b'0000000000' + message.get_bytes())  # 10 more bytes to account for timestamp + type
    if verbose:
        print(message2.to_string())
    assert str(message2.to_string()) == "[BlockRequest(10, 20, True)]"


if __name__ == "__main__":
    test_block_request(True)



"""
Partial transcript from
https://github.com/n-y-z-o/nyzoVerifier/blob/master/src/main/java/co/nyzo/verifier/HashUtil.java
"""


from hashlib import sha256


class HashUtil:

    @staticmethod
    def double_sha256(data: bytes) -> bytes:
        return sha256(sha256(data).digest()).digest()

    @staticmethod
    def single_sha256(data: bytes) -> bytes:
        if data is None:
            data = b''
        return sha256(data).digest()



"""Nyzo Message types"""

from enum import Enum
# see https:#docs.python.org/3/library/enum.html

"""
MessageType(5) will return a <MessageType.Transaction5: 5>
message_type.name will return a str with the MessageType name, like 'Transaction5'
message_type.value will return an int with the MessageType value, like 5 - same a nyzo getValue()
"""


class MessageType(Enum):

    Invalid0 = 0
    _BootstrapRequest1 = 1  # replaced with BootstrapRequestV2_35
    _BootstrapResponse2 = 2  # replaced with BootstrapResponseV2_36
    NodeJoin3 = 3
    NodeJoinResponse4 = 4
    Transaction5 = 5
    TransactionResponse6 = 6
    PreviousHashRequest7 = 7
    PreviousHashResponse8 = 8
    NewBlock9 = 9
    NewBlockResponse10 = 10
    BlockRequest11 = 11
    BlockResponse12 = 12
    TransactionPoolRequest13 = 13
    TransactionPoolResponse14 = 14
    MeshRequest15 = 15
    MeshResponse16 = 16
    StatusRequest17 = 17
    StatusResponse18 = 18
    BlockVote19 = 19
    BlockVoteResponse20 = 20
    NewVerifierVote21 = 21
    NewVerifierVoteResponse22 = 22
    MissingBlockVoteRequest23 = 23
    MissingBlockVoteResponse24 = 24
    MissingBlockRequest25 = 25
    MissingBlockResponse26 = 26
    TimestampRequest27 = 27
    TimestampResponse28 = 28
    HashVoteOverrideRequest29 = 29
    HashVoteOverrideResponse30 = 30
    ConsensusThresholdOverrideRequest31 = 31
    ConsensusThresholdOverrideResponse32 = 32
    NewVerifierVoteOverrideRequest33 = 33
    NewVerifierVoteOverrideResponse34 = 34
    BootstrapRequestV2_35 = 35
    BootstrapResponseV2_36 = 36
    BlockWithVotesRequest37 = 37
    BlockWithVotesResponse38 = 38

    # test messages
    Ping200 = 200
    PingResponse201 = 201

    # maintenance messages
    UpdateRequest300 = 300  # updates the verifier with the latest code from the Git repository, rebuilds, and restarts
    UpdateResponse301 = 301

    # debugging messages -- these are meant to cause problems to test resiliency or to provide information that is not
    # necessary for normal operation
    BlockRejectionRequest400 = 400  # discards all blocks received for the next 10 seconds
    BlockRejectionResponse401 = 401
    DetachmentRequest402 = 402  # stops producing blocks for two verifier cycles
    DetachmentResponse403 = 403
    UnfrozenBlockPoolPurgeRequest404 = 404  # clears the unfrozen block pool
    UnfrozenBlockPoolPurgeResponse405 = 405
    UnfrozenBlockPoolStatusRequest406 = 406  # gets textual information about the unfrozen block pool
    UnfrozenBlockPoolStatusResponse407 = 407
    MeshStatusRequest408 = 408  # gets textual information about the mesh
    MeshStatusResponse409 = 409
    TogglePauseRequest410 = 410  # pauses/un-pauses verifier
    TogglePauseResponse411 = 411
    ConsensusTallyStatusRequest412 = 412
    ConsensusTallyStatusResponse413 = 413
    NewVerifierTallyStatusRequest414 = 414
    NewVerifierTallyStatusResponse415 = 415
    BlacklistStatusRequest416 = 416
    BlacklistStatusResponse417 = 417

    # bootstrapping messages
    ResetRequest500 = 500   # resets the blockchain   TODO: key this to the local verifier before release
    ResetResponse501 = 501

    # the highest allowable message number is 65535
    IncomingRequest65533 = 65533  # for debugging -- passed to readFromStream by meshListener/meshListenerController
    Error65534 = 65534
    Unknown65535 = 65535





from pynyzo.messageobject import MessageObject
from pynyzo.fieldbytesize import FieldByteSize
import json
import struct


class Transaction(MessageObject):
    """Transaction message"""

    nyzos_in_system = 100000000
    micronyzo_multiplier_ratio = 1000000
    micronyzos_in_system = nyzos_in_system * micronyzo_multiplier_ratio

    type_coin_generation = 0
    type_seed = 1
    type_standard = 2

    __slots__ = ('_type', '_timestamp', '_amount', '_receiver_identifier', '_previous_hash_height',
                 '_previous_block_hash', '_sender_identifier', '_sender_data', '_signature')

    """
    type;                   // 1 byte; 0=coin generation, 1=sender verification
    private long timestamp;              // 8 bytes; 64-bit Unix timestamp of the transaction initiation, in milliseconds
    private long amount;                 // 8 bytes; 64-bit amount in micronyzos
    private byte[] receiverIdentifier;   // 32 bytes (256-bit public key of the recipient)

    // Only included in type-1 and type-2 transactions
    private long previousHashHeight;     // 8 bytes; 64-bit index of the block height of the previous-block hash
    private byte[] previousBlockHash;    // 32 bytes (SHA-256 of a recent block in the chain)
    private byte[] senderIdentifier;     // 32 bytes (256-bit public key of the sender)
    private byte[] senderData;           // up to 32 bytes
    private byte[] signature; 
    """

    def __init__(self, buffer: bytes=None, app_log=None):
        super().__init__(app_log=app_log)
        if buffer is None:
            pass
        else:
            # fromByteBuffer constructor
            # These are the fields contained in all transactions.
            offset = 0
            self._type = struct.unpack(">B", buffer[offset:offset +1])[0]  # Byte
            offset += 1
            self._timestamp = struct.unpack(">Q", buffer[offset:offset +8])[0]  # Long, 8
            offset += 8
            self._amount = struct.unpack(">Q", buffer[offset:offset + 8])[0]  # Long, 8
            offset += 8
            self._receiver_identifier = buffer[offset:offset + FieldByteSize.identifier]
            offset += FieldByteSize.identifier
            self.app_log.debug(f"TX( {self._type}, {self._timestamp}, {self._amount}, "
                               f"{self._receiver_identifier.hex()}")
            if self._type == self.type_coin_generation:
                self.app_log.error("TODO: Transaction.type_coin_generation")
            elif self._type in [self.type_seed, self.type_standard]:
                self._previous_hash_height = struct.unpack(">Q", buffer[offset:offset + 8])[0]  # Long, 8
                offset += 8
                self._previous_block_hash = b''  # TODO: get from BlockManager
                self._sender_identifier = buffer[offset:offset + FieldByteSize.identifier]
                offset += FieldByteSize.identifier
                sender_data_length = min(32, struct.unpack(">B", buffer[offset:offset +1])[0])  # Byte
                offset += 1
                self._sender_data = buffer[offset:offset + sender_data_length]  # TODO: memoryview?
                offset += sender_data_length
                self._signature = buffer[offset:offset + FieldByteSize.signature]
                offset += FieldByteSize.signature
                self.app_log.debug(f"TX( {self._previous_hash_height}, {self._sender_identifier.hex()}, "
                                   f"{sender_data_length}, {self._signature.hex()}")
                # print("offset", offset)
            else:
                self.app_log.warning(f"Unknown Transaction type: {self._type}")



    def get_type(self):
        return self._type

    def get_timestamp(self):
        return self._timestamp

    def get_amount(self):
        return self._amount

    def get_amount_after_fee(self):
        return self._amount - self.get_fee()

    def get_receiver_identifier(self):
        return self._receiver_identifier

    def get_sender_identifier(self):
        return self._sender_identifier

    def get_sender_data(self):
        return self._sender_data

    def get_signature(self):
        return self._signature

    def get_fee(self):
        return (self.get_amount() + 399) / 400

    def get_byte_size(self, for_signing:bool=False):

        size = FieldByteSize.transactionType + FieldByteSize.timestamp + FieldByteSize.transactionAmount \
               + FieldByteSize.identifier

        if self._type in [self.type_seed, self.type_standard]:
            if for_signing:
                size += FieldByteSize.hash  # previous-block hash for signing
            else:
                size += FieldByteSize.blockHeight  # previous-hash height for storage and transmission
            size += FieldByteSize.identifier  # sender identifier
            if for_signing:
                size += FieldByteSize.hash  # sender data hash for signing
            else:
                size += 1 + len(self._sender_data) + FieldByteSize.signature  # length specifier + sender data + transaction signature
        return size

    def get_bytes(self, for_signing: bool=False):
        # TODO
        self.app_log.error('TODO: Transaction.get_bytes')

    def to_json(self) -> str:
        return json.dumps({"message_type": "Transaction", 'value': {
            'type': self._type,'timestamp': self._timestamp, 'amount': self._amount,
            'receiver_identifier': self._receiver_identifier.hex(), 'previous_hash_height': self._previous_hash_height,
            'previous_block_hash': self._previous_block_hash.hex(), 'sender_identifier': self._sender_identifier.hex(),
            'sender_data': self._sender_data.hex(), 'signature': self._signature.hex()}})




"""
Nyzo connection layer.
Raw bytes over sockets, with 4 bytes len prefix.
Adapted from eggpool's bismuth code
"""

import socket
import struct
from time import time
import threading

from pynyzo.helpers import base_app_log
from pynyzo.message import Message
from pynyzo.messagetype import MessageType
from pynyzo.messageobject import MessageObject, EmptyMessageObject

# Logical timeout
LTIMEOUT = 45

__version__ = '0.1.0'


class Connection(object):
    """Connection to a Nyzo Node. Handles auto reconnect when needed"""

    __slots__ = ('app_log', 'ip', 'port', 'verbose', 'sdef', 'stats', 'last_activity', 'command_lock', 'raw')

    def __init__(self, ip: str, port: int=9444, verbose: bool=False, app_log: object=None):
        self.app_log = base_app_log(app_log)
        self.ip = ip
        self.port = port
        self.verbose = verbose
        self.sdef = None
        self.last_activity = 0
        self.command_lock = threading.Lock()
        self.check_connection()

    def check_connection(self):
        """Check connection state and reconnect if needed."""
        if not self.sdef:
            try:
                if self.verbose:
                    self.app_log.info(f"Connecting to {self.ip}:{self.port}")
                self.sdef = socket.socket()
                self.sdef.connect((self.ip, self.port))
                self.last_activity = time()
            except Exception as e:
                self.sdef = None
                raise RuntimeError(f"Connections: {e}")

    def send(self, data, retry=True):
        """Sends data buffer to the peer, appends the len header"""
        self.check_connection()
        try:
            self.sdef.settimeout(LTIMEOUT)
            # Header is len of full message, including header
            header = struct.pack(">I", len(data) + 4)
            # Send buffer in one packet - Do not make 2 calls.
            res = self.sdef.sendall(header + data)
            self.last_activity = time()
            # res is always 0 on linux
            if self.verbose:
                self.app_log.info(f"Sent {len(data) + 4} bytes: {(header + data).hex()}")
            return True
        except Exception as e:
            # send failed, try to reconnect
            # TODO: handle tries #
            self.sdef = None
            if retry:
                if self.verbose:
                    self.app_log.warning(f"Send failed ({e}), trying to reconnect")
                self.check_connection()
            else:
                if self.verbose:
                    self.app_log.warning(f"Send failed ({e}), trying to reconnect")
                return False
            try:
                self.sdef.settimeout(LTIMEOUT)
                # Make sure the packet is sent in one call
                res = self.sdef.sendall(header + data)
                return True
            except Exception as e:
                self.sdef = None
                raise RuntimeError(f"Connections: {e}")

    def receive(self):
        """Wait for an answer, for LTIMEOUT sec."""
        self.check_connection()
        self.sdef.settimeout(LTIMEOUT)
        try:
            # Get header
            data = self.sdef.recv(4)
            if not data:
                raise RuntimeError("Socket EOF")
            length = struct.unpack('>I', data)[0] - 4
            if self.verbose:
                self.app_log.info(f"Receiving {length} announced bytes")
        except socket.timeout as e:
            self.sdef = None
            return ""
        try:
            chunks = []
            bytes_recd = 0
            while bytes_recd < length:
                chunk = self.sdef.recv(min(length - bytes_recd, 2048))
                if not chunk:
                    raise RuntimeError("Socket EOF2")
                chunks.append(chunk)
                bytes_recd = bytes_recd + len(chunk)
            self.last_activity = time()
            buffer = b''.join(chunks)
            return buffer
        except Exception as e:
            """
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            """
            self.sdef = None
            raise RuntimeError("Connections: {}".format(e))

    def command(self, command, options=None):
        """
        Sends a command and return it's raw result. - left over from bis, do not use
        """
        with self.command_lock:
            try:
                self.send(command)
                ret = self.receive()
                return ret
            except Exception as e:
                """
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                print(exc_type, fname, exc_tb.tb_lineno)
                """
                # TODO : better handling of tries and delay between
                if self.verbose:
                    print("Error <{}> sending command, trying to reconnect.".format(e))
                self.check_connection()
                self.send(command)
                ret = self.receive()
                return ret

    def close(self):
        """Close the socket"""
        try:
            self.sdef.close()
        except Exception as e:
            pass

    def fetch_buffer(self, message: Message, identifier: bytes=b'') -> bytes:
        """From Message - fetch bin buffer"""
        # TODO
        # identifier = NodeManager.identifierForIpAddress(self.ip);
        self.send(message.get_bytes_for_transmission())
        response = self.receive()
        return response

    def fetch(self, message: Message, identifier: bytes=b'') -> MessageObject:
        """Fetch then decode"""
        buffer = self.fetch_buffer(message, identifier)
        return Message.from_bytes(buffer, b'').get_content()  # self.ip, but convert or use str for ips?


if __name__ == "__main__":
    print("I'm a module, can't run!")



import pytest
import sys
import ed25519
import hashlib

sys.path.append('../')
from pynyzo.keyutil import KeyUtil
from pynyzo.byteutil import ByteUtil


def test_keys_1(verbose=False):
    master = b'9444'  # Nyzo port as test seed
    seed = hashlib.sha256(master).digest()
    assert seed == b',V@?\xcf\xfb\xd14M\xfbw&\x89\xf6\x0eDoDq\xd2\xca\x17\xf3\xa5\xea\xa6)\xf6\xc52<l'
    # NEVER USE THAT KEY IN REAL WORLD!!!
    signing_key = ed25519.SigningKey(seed)
    hex = signing_key.to_ascii(encoding="hex").decode('utf-8')
    if verbose:
        print("private", hex)
    assert hex == '2c56403fcffbd1344dfb772689f60e446f4471d2ca17f3a5eaa629f6c5323c6c'
    public = KeyUtil.private_to_public(hex)
    if verbose:
        print("public", public)
    assert public == 'a34abe2176eeb545735b25ff68c41d716db462e55c7c3a5d48db0ae4fc95ae9e'


def test_keys_2(verbose=False):
    master = b'9444'  # Nyzo port as test seed
    seed = hashlib.sha256(master).digest()
    # NEVER USE THAT KEY IN REAL WORLD!!!
    signing_key = ed25519.SigningKey(seed)
    nyzo_format = ByteUtil.bytes_as_string_with_dashes(signing_key.to_bytes()[:32])
    if verbose:
        print("private nyzo format", nyzo_format)
    assert nyzo_format == '2c56403f-cffbd134-4dfb7726-89f60e44-6f4471d2-ca17f3a5-eaa629f6-c5323c6c'


def test_sign_1(verbose=False):
    master = b'9444'  # Nyzo port as test seed
    seed = hashlib.sha256(master).digest()
    # NEVER USE THAT KEY IN REAL WORLD!!!
    signing_key = ed25519.SigningKey(seed)
    message = b'Message to sign'
    sig = signing_key.sign(message)
    if verbose:
        print("Signature", sig.hex())
    assert sig.hex() == 'df5cc085f0892837404525b9524089e4ecf703e050c0d42b798d687356c3984c9f86c0958353f861a17cf8d6a0019cc90a2c8414021fd64110ce0e6c5a923f00'


def test_verify_1(verbose=False):
    message = b'Message to sign'
    public = bytes.fromhex('a34abe2176eeb545735b25ff68c41d716db462e55c7c3a5d48db0ae4fc95ae9e')
    signature = bytes.fromhex('df5cc085f0892837404525b9524089e4ecf703e050c0d42b798d687356c3984c9f86c0958353f861a17cf8d6a0019cc90a2c8414021fd64110ce0e6c5a923f00')
    verifying_key = ed25519.VerifyingKey(public)
    try:
        verifying_key.verify(signature, message)
        if verbose:
            print("signature is good")
    except:
        raise

if __name__ == "__main__":
    # test_keys_2(True)
    # KeyUtil.main()
    test_sign_1(True)
    test_verify_1(True)



"""
Mostly transcripted from
https://github.com/n-y-z-o/nyzoVerifier/blob/master/src/main/java/co/nyzo/verifier/MessageObject.java
"""

from abc import ABC, abstractmethod
from pynyzo.helpers import base_app_log
import json


class MessageObject(ABC):

    __slots__ = ('app_log', )

    def __init__(self, app_log: object=None):
        self.app_log = base_app_log(app_log)

    @abstractmethod
    def get_byte_size(self) -> int:
        pass

    @abstractmethod
    def get_bytes(self) -> bytes:
        pass

    def to_json(self) -> str:
        return json.dumps({"Error": "Not implemented"})


class EmptyMessageObject(MessageObject):
    """This one was added for convenience and avoid extra tests, not present in the java impl."""

    def get_byte_size(self) -> int:
        return 0

    def get_bytes(self) -> bytes:
        return b''



"""
From https://github.com/n-y-z-o/nyzoVerifier/blob/master/src/main/java/co/nyzo/verifier/FieldByteSize.java
"""


class FieldByteSize:

    balanceListLength = 4
    blockHeight = 8
    blocksUntilFee = 2
    booleanField = 1
    cycleLength = 4
    frozenBlockListLength = 2
    hash = 32
    hashListLength = 1
    ipAddress = 4
    messageLength = 4
    port = 4
    rolloverTransactionFees = 1
    seed = 32
    timestamp = 8
    transactionAmount = 8
    transactionType = 1
    messageType = 2
    identifier = 32
    nodeListLength = 4
    signature = 64
    stringLength = 2
    transactionPoolLength = 4
    unfrozenBlockPoolLength = 2
    unnamedDouble = 8
    unnamedInteger = 4
    unnamedShort = 2
    voteListLength = 1

    @staticmethod
    def string(value: str) -> int:
        return FieldByteSize.stringLength + len(value.encode('utf-8'))





from pynyzo.messageobject import MessageObject
from pynyzo.helpers import base_app_log
from pynyzo.balancelistitem import BalanceListItem
from pynyzo.fieldbytesize import FieldByteSize
from pynyzo.hashutil import HashUtil
import json
import struct
import sys


class BalanceList(MessageObject):
    """BalanceList message"""

    __slots__ = ('_block_height', '_rollover_fees', '_previous_verifiers', '_items')

    def __init__(self, block_height: int=0, rollover_fees: int=0, previous_verifiers: list=None, items: list=None,
                 buffer: bytes=None, app_log=None):
        # This replaces the various constructors from java, depending on the params
        super().__init__(app_log=app_log)
        if buffer:
            # buffer is the full buffer with timestamp and type, why the 10 offset.
            offset = 0
            self._block_height = struct.unpack(">Q", buffer[offset:offset + 8])[0]  # long, 8
            offset += 8
            self._rollover_fees = struct.unpack(">B", buffer[offset:offset + 1])[0]  # byte
            offset += 1
            number_of_previous_verifiers = min(self._block_height, 9)
            self.app_log.debug(f"BalanceList({self._block_height}, {self._rollover_fees}, {number_of_previous_verifiers})")
            self._previous_verifiers = []
            for i in range(number_of_previous_verifiers):
                # We could use a memoryview if perf /ram was an issue
                self._previous_verifiers.append(buffer[offset:offset + FieldByteSize.identifier])
                offset += FieldByteSize.identifier
            number_of_pairs = struct.unpack(">I", buffer[offset:offset + 4])[0]  # int, 4
            offset += 4
            # print("number_of_pairs", number_of_pairs)
            self._items = []
            for i in range(number_of_pairs):
                identifier = buffer[offset:offset + FieldByteSize.identifier]
                offset += FieldByteSize.identifier
                balance = struct.unpack(">Q", buffer[offset:offset + 8])[0]  # long, 8
                offset += 8
                blocks_until_fee = struct.unpack(">H", buffer[offset:offset + 2])[0]  # Short, 2 bytes
                offset += 2
                item = BalanceListItem(identifier, balance, blocks_until_fee)
                # print(item.to_json())
                self._items.append(item)
        else:
            self._block_height = block_height
            self._rollover_fees = rollover_fees
            self._previous_verifiers = previous_verifiers
            self._items = items

    def get_block_height(self) -> int:
        return self._block_height

    def get_rollover_fees(self) -> int:
        return self._rollover_fees

    def get_previous_verifiers(self) -> list:
        return self._previous_verifiers.copy()  # shallow copy is enough for that type.

    def get_items(self) -> list:
        return self._items

    def get_byte_size(self) -> int:
        number_of_previous_verifiers = min(self._block_height, 9)
        bytes_per_item = FieldByteSize.identifier + FieldByteSize.transactionAmount + FieldByteSize.blocksUntilFee

        return FieldByteSize.blockHeight + FieldByteSize.rolloverTransactionFees \
               + FieldByteSize.identifier * number_of_previous_verifiers + FieldByteSize.balanceListLength \
               + bytes_per_item * len(self._items)

    def get_bytes(self) -> bytes:
        result = b''
        self.app_log.error("TODO: BalanceList.get_bytes")
        return result

    def get_hash(self) -> bytes:
        return HashUtil.double_sha256(self.get_bytes())

    def to_string(self) -> str:
        return f"[BalanceList: height={self._block_height}, hash={self.get_hash().hex()}]"

    def to_json(self) -> str:
        # Do not add explicit keys for balance items, too verbose.
        items = {item.get_identifier().hex(): [item.get_balance(), item.get_blocks_until_fee()] for item in self._items}
        previous_verifiers = [verifier.hex() for verifier in self._previous_verifiers]
        return json.dumps({"message_type": "BalanceList", 'value': {
            'height': self._block_height, 'rollover_fees': self._rollover_fees,
            'previous_verifiers': previous_verifiers, "items": items}})







"""
Message ancestor class for Nyzo messages
"""

import struct
from abc import ABC, abstractmethod
from pynyzo.helpers import base_app_log
from pynyzo.messagetype import MessageType
from pynyzo.messageobject import MessageObject, EmptyMessageObject
from pynyzo.fieldbytesize import FieldByteSize
from pynyzo.keyutil import KeyUtil
from pynyzo.byteutil import ByteUtil
import pynyzo.config as config
from pynyzo.messages.statusresponse import StatusResponse
from pynyzo.messages.blockresponse import BlockResponse

from time import time

__version__ = '0.0.1'


class Message(ABC):
    """Abstract Ancestor for all messages."""

    # slots for private vars, to spare ram
    __slots__ = ('_applog', '_timestamp', '_type', '_content', '_sourceNodeIdentifier', '_sourceNodeSignature',
                 '_valid', '_source_ip_address')

    # Static class variables
    maximumMessageLength = 4194304  # 4MB
    whitelist = dict()
    disallowedNonCycleTypes = (MessageType.NewBlock9, MessageType.BlockVote19, MessageType.NewVerifierVote21,
                               MessageType.MissingBlockVoteRequest23, MessageType.MissingBlockRequest25)

    # We do not broadcast any messages to the full mesh from the broadcast method. We do, however, use the full mesh
    # as a potential pool for random requests for the following types. This reduces strain on in-cycle verifiers.
    fullMeshMessageTypes = (MessageType.BlockRequest11, MessageType.BlockWithVotesRequest37)

    def __init__(self, a_type: MessageType, content: MessageObject, app_log:object=None,
                 sourceNodeIdentifier: bytes=None, sourceNodeSignature: bytes=None, source_ip_address: bytes=None,
                 timestamp: int=0):
        """This is the constructor for a new message originating from this system AND from the outside,
        depending on the params."""
        self.app_log = base_app_log(app_log)
        self._type = a_type
        self._content = content
        self._valid = True

        if sourceNodeIdentifier is None:
            # From our system
            self._timestamp = int(time()*1000)
            self._sourceNodeIdentifier = config.PUBLIC_KEY.to_bytes()
            self._sourceNodeSignature = KeyUtil.sign_bytes(self.get_bytes_for_signing(), config.PRIVATE_KEY)
            # As a test, we can verify our sig before sending
            """
            KeyUtil.signature_is_valid(self._sourceNodeSignature, self.get_bytes_for_signing(), 
                                       config.PUBLIC_KEY.to_bytes())
            """
        else:
            # From another system
            self._timestamp = timestamp
            self._sourceNodeIdentifier = sourceNodeIdentifier
            self._sourceNodeSignature = sourceNodeSignature
            self._sourceIpAddress = source_ip_address
            # self._valid = KeyUtil.signature_is_valid(sourceNodeSignature, self.get_bytes_for_signing(),
            # sourceNodeIdentifier)
            self.valid = True
            # TODO: needs all the chain of get_bytes to validate; let suppose for now it is valid.
            self.app_log.warning(f"TODO: Did NOT validate message from "
                                 f"{ByteUtil.bytes_as_string_with_dashes(sourceNodeIdentifier)} "
                                 f"of type {self._type.name}")
            if not self._valid:
                self.app_log.warning(f"message from {ByteUtil.bytes_as_string_with_dashes(sourceNodeIdentifier)} "
                                     f"of type {self._type.name} is not valid")  # Temp log

    def to_string(self) -> str:
        """String view of the message for log/print"""
        return f"[Message: {self._type.name} ({self.get_content()})]"

    def get_timestamp(self) -> int:
        return self._timestamp

    def get_type(self) -> MessageType:
        return self._type

    def get_content(self) -> MessageObject:
        return self._content

    def get_source_node_identifier(self) -> bytes:
        return self._sourceNodeIdentifier

    def get_source_node_signature(self) -> bytes:
        return self._sourceNodeSignature

    def is_valid(self) -> bool:
        return self._valid

    def get_source_ip_address(self) -> bytes:
        return self._sourceIpAddress

    def sign(self, private_seed: bytes) -> None:
        # TODO
        # self.app_log.error("TODO: Message.sign()")
        self._sourceNodeIdentifier = KeyUtil.private_to_public(private_seed.hex())
        self._sourceNodeSignature = KeyUtil.signBytes(self.get_bytes_for_signing(), private_seed)

    def get_bytes_for_transmission(self) -> bytes:
        """"""
        # Determine the size (timestamp, type, source-node identifier, source-node signature, content if present).
        size_bytes = FieldByteSize.messageLength + FieldByteSize.timestamp + FieldByteSize.messageType \
                     + FieldByteSize.identifier + FieldByteSize.signature
        size_bytes += self._content.get_byte_size()

        # Make the buffer.
        buffer = b''
        # Header is low level, added when sending
        # buffer += struct.pack('I', size_bytes)  # 4 bytes

        # Add the data.
        buffer += struct.pack('>Q', self._timestamp)   # unsigned long long, 8 bytes
        self.app_log.debug(f"get bytes for message {self._type.name}, {self._type.value}")
        buffer += struct.pack('>h', self._type.value)  # 2 bytes
        buffer += self._content.get_bytes()  # no need for test, see EmptyMessageObject
        buffer += self._sourceNodeIdentifier
        buffer += self._sourceNodeSignature
        self.app_log.debug(f"buffer ({len(buffer)}): {buffer.hex()}")
        return buffer

    def get_bytes_for_signing(self) -> bytes:
        """"""
        # Determine the size (timestamp, type, source-node identifier, content if present).
        size_bytes = FieldByteSize.timestamp + FieldByteSize.messageType + FieldByteSize.identifier
        size_bytes += self._content.get_byte_size()

        # Make the buffer.
        buffer = b''
        # Header is low level, added when sending
        # buffer += struct.pack('I', size_bytes)  # 4 bytes

        # Add the data.
        buffer += struct.pack('>Q', self._timestamp)   # unsigned long long, 8 bytes
        self.app_log.debug(f"get bytes for signing {self._type.name}, {self._type.value}")
        buffer += struct.pack('>h', self._type.value)  # 2 bytes
        buffer += self._content.get_bytes()  # no need for test, see EmptyMessageObject
        buffer += self._sourceNodeIdentifier
        self.app_log.debug(f"buffer to sign ({len(buffer)}): {buffer.hex()}")
        return buffer

    @staticmethod
    def from_bytes(buffer: bytes, source_ip_address: bytes) -> any:
        message = None
        message_type = None

        timestamp = struct.unpack('>Q', buffer[:8])[0]
        typeValue = struct.unpack('>h', buffer[8:10])[0]
        message_type = MessageType(typeValue)

        content = Message.process_content(message_type, buffer)

        # size = len(buffer)
        sourceNodeIdentifier = buffer[-64-32:-64]
        sourceNodeSignature = buffer[-64:]  # sig is last 64 bytes

        # print("id", sourceNodeIdentifier.hex(), "sig", sourceNodeSignature.hex() )

        message = Message(timestamp=timestamp, a_type=message_type, content=content,
                          sourceNodeIdentifier=sourceNodeIdentifier, sourceNodeSignature=sourceNodeSignature,
                          source_ip_address=source_ip_address)
        return message

    @staticmethod
    def process_content(message_type: MessageType, buffer: bytes) -> MessageObject:
        content = EmptyMessageObject()
        if message_type == MessageType.StatusResponse18:
            content = StatusResponse(buffer=buffer)
        if message_type == MessageType.BlockResponse12:
            content = BlockResponse(buffer=buffer)
        return content



"""
Helper Class and functions
"""

import logging
import struct
# Using tornado.log for pretty printing. async client could follow.
import tornado.log
from os import path
from logging.handlers import RotatingFileHandler


LOG_LEVEL = 'DEBUG'


def base_app_log(app_log=None):
    """Returns the best possible log handler if none is provided"""
    if app_log:
        return app_log
    elif logging.getLogger("tornado.application"):
        return logging.getLogger("tornado.application")
    else:
        return logging


def tornado_logger():
    """Define a tornado logger"""
    app_log = logging.getLogger("tornado.application")
    app_log.setLevel(LOG_LEVEL)
    tornado.log.enable_pretty_logging()
    logfile = path.abspath("nyzo.log")
    # Rotate log after reaching 512K, keep 5 old copies.
    rotate_handler = RotatingFileHandler(logfile, "a", 512 * 1024, 5)
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    rotate_handler.setFormatter(formatter)
    app_log.addHandler(rotate_handler)
    return app_log


def strings_to_buffer(lines: list) -> bytes:
    """Diverges from Nyzo arch"""
    result = []
    result.append(struct.pack(">B", len(lines)))  # byte
    for line in lines:
        bin = line.encode('utf-8')
        result.append(struct.pack(">h", len(bin)))  # short
        result.append(bin)
    return b''.join(result)


def buffer_to_strings(b: memoryview) -> list:
    number_of_lines = struct.unpack(">B", b[:1])[0]
    result = list()
    pos = 1
    for i in range(number_of_lines):
        line_len = struct.unpack(">h", b[pos:pos + 2])[0]
        result.append(bytes(b[pos+2:pos+2+line_len]).decode('utf-8'))
        pos += line_len + 2
    return result



"""

"""


def split_by_len(item, maxlen):
    return (item[ind:ind+maxlen] for ind in range(0, len(item), maxlen))


class ByteUtil:

    @staticmethod
    def bytes_as_string_with_dashes(buffer: bytes) -> str:
        hexa = buffer.hex()
        return '-'.join(split_by_len(hexa, 8))



"""
PyNyzo
Common variables

Also serves as config file for POC and tests
"""

from os import path
from pynyzo.keyutil import KeyUtil
from pynyzo.byteutil import ByteUtil

__version__ = '0.0.1'

"""
User config - You can change these ones, but **See doc**
Please use config.txt instead of editing here.  

This file is only to be edited by dev or for debug purposes.
"""

VERBOSE = True

# Dev only - break nice color
DEBUG = False

# Dev only, dumps network packets
DUMP_PACKETS = False

# optional log details we may want. AVAILABLE_LOGS lists all possible extra details.
# Copy the ones you want in LOG
LOG = []

AVAILABLE_LOGS = ['keys', 'connections', 'timing']

# Path to the privkey
NYZO_SEED = "tmp/verifier_private_seed"

PYTHON_EXECUTABLE = "python3"


"""
Here comes tuneable algorithm variables - Do not change those or you may  be unable to sync
"""

# Default Nyzo port
DEFAULT_PORT = 9444

# How long to wait before retrying a failed peer?
PEER_RETRY_SECONDS = 20

# Seconds between pings
PING_DELAY = 30

"""
DEBUG VARS
"""

NETWORK_TIMEOUT = 300

"""
Global Variables
"""

PRIVATE_KEY = b''
PUBLIC_KEY = b''

# --------------------------------------------------- #

# The potential variables and their type
_VARS = {
    "VERBOSE": "bool",
    "DEBUG": "bool",
    "DUMP_PACKETS": "bool",
    "LOG": "list",
    "NYZO_SEED": "str",
    "PYTHON_EXECUTABLE": "str",
    "PEER_RETRY_SECONDS": "int",
    "PING_DELAY": "int",
    "NETWORK_TIMEOUT": "int"
}


def _get_left_right(line, var_list):
    left, right = map(str.strip, line.rstrip("\n").split("="))
    if left not in var_list:
        return None, None
    param = var_list[left]
    if param == "int":
        right = int(right)
    elif param == "float":
        right = float(right)
    elif param == "list":
        # Separator may be ; or ,
        right = [str(item.strip()) for item in right.replace(';', ',').split(",")]
    elif param == "bool":
        if right.lower() in ["false", "0", "", "no"]:
            right = False
        else:
            right = True
    else:
        # treat as "str"
        pass
    return left, right


def _overload(file_name: str):
    for line in open(file_name):
        if line and line[0] == '#':
            continue
        if '=' in line:
            left, right = _get_left_right(line, _VARS)
            if left:
                globals().update({left: right})


def load(dir: str='./'):
    """
    Overload info with config.default.txt and config.txt
    :return:
    """
    global PRIVATE_KEY
    global PUBLIC_KEY
    if path.exists(dir + "config.default.txt"):
        _overload(dir + "config.default.txt")
    if path.exists(dir + "config.txt"):
        _overload(dir + "config.txt")

    # Load the keys
    if not path.isfile(NYZO_SEED):
        print(f"No key file, creating one into {NYZO_SEED}")
        PRIVATE_KEY = KeyUtil.generateSeed()
        KeyUtil.save_to_private_seed_file(NYZO_SEED, PRIVATE_KEY)
    PRIVATE_KEY, PUBLIC_KEY = KeyUtil.get_from_private_seed_file(NYZO_SEED)
    # We can tweak verbosity later on, do not print here but later on.
    if DEBUG:
        print(f"Key Loaded, public id {ByteUtil.bytes_as_string_with_dashes(PUBLIC_KEY.to_bytes())}")

    return {var: globals()[var] for var in _VARS}


if __name__ == "__main__":
    print("I'm a module, can't run!")





from pynyzo.messageobject import MessageObject
from pynyzo.fieldbytesize import FieldByteSize
import json
import struct


class BalanceListItem(MessageObject):
    """BalanceListItem message"""

    transfer_identifier = bytes.fromhex("0000000000000000000000000000000000000000000000000000000000000001")

    __slots__ = ('_blocks_betweenfee', '_identifier', '_balance', '_blocks_until_fee')

    def __init__(self, identifier: bytes=None, balance: int=0, blocks_until_fee: int=None, buffer: bytes = None,
                 app_log=None):
        # This replaces the various constructors from java, depending on the params
        super().__init__(app_log=app_log)
        self._blocks_betweenfee = 500
        if buffer is None:
            self._identifier = identifier
            self._balance = balance
            if blocks_until_fee == None:
                self._blocks_until_fee = 0 if identifier == self.transfer_identifier else self._blocks_betweenfee
            else:
                self._blocks_until_fee = blocks_until_fee
        else:
            pass

    def get_identifier(self) -> bytes:
        return self._identifier

    def get_balance(self) -> int:
        return self._balance

    def get_blocks_until_fee(self) -> int:
        return self._blocks_until_fee

    def get_byte_size(self) -> int:
        return FieldByteSize.identifier + 8 + 2

    def get_bytes(self) -> bytes:
        self.app_log.error("TODO: BalanceListItem.get_bytes")
        return b''

    def to_json(self) -> str:
        # I stripped one level, type and value
        return json.dumps({'identifier': self._identifier.hex(), 'balance': self._balance,
                           'blocks_until_fee': self._blocks_until_fee})




"""
Eddsa Ed25519 key handling
From
https://github.com/n-y-z-o/nyzoVerifier/blob/b73bc25ba3094abe3470ec070ce306885ad9a18f/src/main/java/co/nyzo/verifier/KeyUtil.java

plus
https://github.com/n-y-z-o/nyzoVerifier/blob/17509f03a7f530c0431ce85377db9b35688c078e/src/main/java/co/nyzo/verifier/util/SignatureUtil.java
"""

# Uses https://github.com/warner/python-ed25519 , c binding, fast


import ed25519
import hashlib
from pynyzo.byteutil import ByteUtil


class KeyUtil:

    @staticmethod
    def main():
        """Temp test, not to be used"""
        signing_key, verifying_key = ed25519.create_keypair()
        print("Original private key", ByteUtil.bytes_as_string_with_dashes(signing_key.to_bytes()[:32]))
        # signing key has signing + verifying, we keep the first 32 to only get the private part.
        print("Original public key", ByteUtil.bytes_as_string_with_dashes(verifying_key.to_bytes()))

    @staticmethod
    def generateSeed(hashable_keyword: str='') -> bytes:
        """Generate a private key, with optional keyword to get reproducible tests results or later HD Wallet."""
        if len(hashable_keyword):
            seed = hashlib.sha256(hashable_keyword).digest()
            signing_key = ed25519.SigningKey(seed)
        else:
            signing_key, _ = ed25519.create_keypair()
        return signing_key.to_bytes()[:32]

    @staticmethod
    def private_to_public(private: str) -> str:
        """Temp Test"""
        keydata = bytes.fromhex(private)
        signing_key = ed25519.SigningKey(keydata)
        verifying_key = signing_key.get_verifying_key()
        vkey_hex = verifying_key.to_ascii(encoding="hex")
        return vkey_hex.decode('utf-8')

    @staticmethod
    def get_from_private_seed_file(filename: str):
        """returns priv and pub key - as object - from the stored nyzo text id format"""
        with open(filename) as f:
            nyzo = f.read(80).replace('-', '').encode('utf-8').strip()
            signing_key = ed25519.SigningKey(nyzo, encoding="hex")
            verifying_key = signing_key.get_verifying_key()
        return signing_key, verifying_key

    @staticmethod
    def save_to_private_seed_file(filename: str, key: bytes) -> None:
        """Saves the privkey to the nyzo formatted file"""
        nyzo_format = ByteUtil.bytes_as_string_with_dashes(key)
        with open(filename, 'w') as f:
            f.write(nyzo_format)

    @staticmethod
    def sign_bytes(bytes_to_sign: bytes, private_key: ed25519.SigningKey) -> bytes:
        sig = private_key.sign(bytes_to_sign)
        return sig

    @staticmethod
    def signature_is_valid(signature: bytes, signed_bytes: bytes, public_id: bytes) -> bool:
        verifying_key = ed25519.VerifyingKey(public_id)
        # todo: cache key from id, see https://github.com/n-y-z-o/nyzoVerifier/blob/17509f03a7f530c0431ce85377db9b35688c078e/src/main/java/co/nyzo/verifier/util/SignatureUtil.java
        try:
            verifying_key.verify(signature, signed_bytes)
            # print("signature is good")
            return True
        except ed25519.BadSignatureError:
            # print("signature is bad!")
            return False


if __name__ == "__main__":
    KeyUtil.main()
    # KeyUtil.private_to_public('nyzo-formatted-private-key'.replace('-', ''))





"""
NYZO
Get status test script
"""

import argparse
import sys

sys.path.append('../')
from pynyzo.messagetype import MessageType
from pynyzo.message import Message
from pynyzo.messageobject import EmptyMessageObject
# from pynyzo.messages.statusresponse import StatusResponse
from pynyzo.messages.blockrequest import BlockRequest
from pynyzo.connection import Connection
from pynyzo.helpers import tornado_logger
import pynyzo.config as config


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Nyzo test')
    parser.add_argument("-I", "--ip", type=str, default='127.0.0.1', help="IP to query (default 127.0.0.1)")
    parser.add_argument("-v", "--verbose", action="count", default=False, help='Be verbose.')
    parser.add_argument("-a", "--action", type=str, default='status', help='Action (status, block)')
    args = parser.parse_args()

    app_log = tornado_logger()

    config.load()

    connection = Connection(args.ip, app_log=app_log, verbose=args.verbose)

    if args.action == 'status':
        empty = EmptyMessageObject(app_log=app_log)
        message = Message(MessageType.StatusRequest17, empty, app_log=app_log)
        res = connection.fetch(message)
        print(res.to_json())
    elif args.action == 'block':
        # test. Use a block high enough so it's not frozen.
        # TODO: get last block from  reference verifiers
        request = BlockRequest(start_height=1695400, end_height=1695400+2, include_balance_list=False, app_log=app_log)
        message = Message(MessageType.BlockRequest11, request, app_log=app_log)
        res = connection.fetch(message)
        print(res.to_json())





import pytest
import sys
import ed25519
import hashlib

sys.path.append('../')
from pynyzo.hashutil import HashUtil

# See https://github.com/n-y-z-o/nyzoVerifier/blob/master/src/main/java/co/nyzo/verifier/HashUtil.java


def test_single_1(verbose=False):
    hello = b"hello"
    hex = HashUtil.single_sha256(hello).hex()
    if verbose:
        print("hex", hex)
    assert hex == '2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824'


def test_single_2(verbose=False):
    hello = b''
    hex = HashUtil.single_sha256(hello).hex()
    if verbose:
        print("hex", hex)
    assert hex == 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'


def test_double_1(verbose=False):
    hello = b"hello"
    hex = HashUtil.double_sha256(hello).hex()
    if verbose:
        print("hex", hex)
    assert hex == '9595c9df90075148eb06860365df33584b75bff782a510c6cd4883a419833d50'


def test_double_2(verbose=False):
    hello = b''
    hex = HashUtil.double_sha256(hello).hex()
    if verbose:
        print("hex", hex)
    assert hex == '5df6e0e2761359d30a8275058e299fcc0381534545f55cf43e41983f5d4c9456'


if __name__ == "__main__":
    test_single_1(True)
    test_single_2(True)
    test_double_1(True)
    test_double_2(True)





from pynyzo.messageobject import MessageObject
from pynyzo.fieldbytesize import FieldByteSize
from pynyzo.transaction import Transaction
import json
import struct

from enum import Enum


class Block(MessageObject):
    """Block message"""

    class ContinuityState(Enum):
        Undetermined = 0
        Discontinuous = 1
        Continuous = 2

    class SignatureState(Enum):
        Undetermined = 0
        Valid = 1
        Invalid = 2

    __slots__ = ('_height', '_previous_block_hash', '_start_timestamp', '_verification_timestamp', '_transactions',
                 '_balance_list_hash', '_verifier_identifier', '_verifier_signature', '_continuity_state',
                 '_signature_state', '_cycle_information')
    """
    private long height;                           // 8 bytes; 64-bit integer block height from the Genesis block,
                                                   // which has a height of 0
    private byte[] previousBlockHash;              // 32 bytes (this is the double-SHA-256 of the previous block
                                                   // signature)
    private long startTimestamp;                   // 8 bytes; 64-bit Unix timestamp of the start of the block, in
                                                   // milliseconds
    private long verificationTimestamp;            // 8 bytes; 64-bit Unix timestamp of when the verifier creates the
                                                   // block, in milliseconds
    private List<Transaction> transactions;        // 4 bytes for number + variable
    private byte[] balanceListHash;                // 32 bytes (this is the double-SHA-256 of the account balance list)
    private byte[] verifierIdentifier;             // 32 bytes
    private byte[] verifierSignature;              // 64 bytes

    private ContinuityState continuityState = ContinuityState.Undetermined;
    private SignatureState signatureState = SignatureState.Undetermined;
    private CycleInformation cycleInformation = null;
    """

    def __init__(self, height: int=0, previous_block_hash: bytes=None, start_timestamp: int=0, transactions:list=None,
                 balance_list_hash:bytes=None, buffer: bytes=None, offset=0, app_log=None):
        super().__init__(app_log=app_log)
        if buffer is None:
            # TODO
            pass
        else:
            # Same as original fromByteBuffer constructor
            self._height = struct.unpack(">Q", buffer[offset:offset +8])[0]  # Long, 8
            offset += 8
            self._previous_block_hash = buffer[offset:offset +32]
            offset += 32
            self._start_timestamp = struct.unpack(">Q", buffer[offset:offset +8])[0]  # Long, 8
            offset += 8
            self._verification_timestamp = struct.unpack(">Q", buffer[offset:offset +8])[0]  # Long, 8
            offset += 8
            number_of_transactions = struct.unpack(">I", buffer[offset:offset +4])[0]  # Int, 4
            offset += 4
            self._transactions = []
            self.app_log.debug(f"Block {self._height}, {self._previous_block_hash.hex()}, {self._start_timestamp}, "
                               f"{self._verification_timestamp}, {number_of_transactions}")
            mv = memoryview(buffer)
            for i in range(number_of_transactions):
                transaction = Transaction(buffer=mv[offset:])
                added = transaction.get_byte_size()
                offset += added
                self._transactions.append(transaction)
            self._balance_list_hash = buffer[offset:offset + FieldByteSize.hash]
            offset += FieldByteSize.hash
            self._verifier_identifier = buffer[offset:offset + FieldByteSize.identifier]
            offset += FieldByteSize.identifier
            self._verifier_signature = buffer[offset:offset + FieldByteSize.signature]
            offset += FieldByteSize.signature
            # print("block offset", offset)

    def get_bytes(self, include_signature: bool=False) -> bytes:
        # TODO
        self.app_log.error('TODO: Block.get_bytes')
        """
        int size = getByteSize();

        // Assemble the buffer.
        byte[] array = new byte[size];
        ByteBuffer buffer = ByteBuffer.wrap(array);
        buffer.putLong(height);
        buffer.put(previousBlockHash);
        buffer.putLong(startTimestamp);
        buffer.putLong(verificationTimestamp);
        buffer.putInt(transactions.size());
        for (Transaction transaction : transactions) {
            buffer.put(transaction.getBytes());
        }
        buffer.put(balanceListHash);
        if (includeSignature) {
            buffer.put(verifierIdentifier);
            buffer.put(verifierSignature);
        }

        return array;
        """

    def get_byte_size(self, include_signature: bool=False) -> int:
        size = FieldByteSize.blockHeight + FieldByteSize.hash + FieldByteSize.timestamp  + FieldByteSize.timestamp \
               + 4 + FieldByteSize.hash
        for transaction in self._transactions:
            size += transaction.get_byte_size()
        if include_signature:
            size += FieldByteSize.identifier + FieldByteSize.signature
        # print("block size", size)
        return size

    def to_json(self) -> str:
        transactions = [json.loads(tx.to_json()) for tx in self._transactions]
        return json.dumps({"message_type": "Block", 'value': {
            'height': self._height, 'previous_block_hash': self._previous_block_hash.hex(),
            'start_timestamp': self._start_timestamp, 'verification_timestamp': self._verification_timestamp,
            'transactions': transactions, 'balance_list_hash': self._balance_list_hash.hex(),
            'verifier_identifier': self._verifier_identifier.hex(),
            'verifier_signature': self._verifier_signature.hex()}})
